/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - server.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

/* eslint-disable no-console */
import {APP_BASE_HREF} from "@angular/common";
import {environment} from "@environments/environment";
import {ngExpressEngine} from "@nguniversal/express-engine";
import * as express from "express";
import {existsSync} from "fs";
import {join} from "path";
import {
  InMemoryStorage,
  isNullOrUndefined,
  SsrUtil,
} from "solidify-frontend";
import "zone.js/node";

import {AppServerModule} from "./src/main.server";

const DEFAULT_SHOW_LOGS: boolean = false;
const DEFAULT_LOG_TIME: boolean = false;

const PROJECT_NAME: string = "my-app-portal"; /* TODO DEFINE APP NAME HERE */
const DEFAULT_SSR_PORT: number = 4200; /* TODO DEFINE PORT HERE */

const {createProxyMiddleware} = require("http-proxy-middleware");
const proxyConfLocal = require("./proxy.conf.local.js");

const domino = require("domino");
const fs = require("fs");
const path = require("path");
const template = fs.readFileSync(path.join(".", "dist", PROJECT_NAME, "browser", "index.html")).toString();

// Shim for the global window and document objects.
const window = domino.createWindow(template);
window.localStorage = new InMemoryStorage();
window.sessionStorage = new InMemoryStorage();
window.navigator.language = "fr-CH";
window.navigator.onLine = true;
window.encodeURIComponent = (v: string) => v;
window.decodeURIComponent = (v: string) => v;
window[SsrUtil.SSR_PROPERTY_KEY] = true;
window[SsrUtil.IS_SSR_FIRST_CALL_PROPERTY_KEY] = true;
global[SsrUtil.WINDOW_PROPERTY_KEY] = window;
(SsrUtil as any)[SsrUtil.WINDOW_PROPERTY_KEY] = window;

const ssrRoutes: string[] = [
  "/",
  "/home",
  "/about",
];

const staticRoutes: string[] = [
  "./**",
  "./assets/**",
  "assets/**",
  "/assets/**",
  "*.*",
];

const log: (...otherParam: any[]) => void = (...otherParam: any[]) => {
  if (environment.showDebugInformation) {
    console.log(...otherParam);
  }
};

// The Express app is exported so that it can be used by serverless Functions.
export const app = (showLogs: boolean, logTime: boolean): express.Express => {
  const server = express();
  const distFolder = join(process.cwd(), "dist/" + PROJECT_NAME + "/browser");
  const indexHtml = existsSync(join(distFolder, "index.original.html")) ? "index.original.html" : "index";

  // Our Universal express-engine (found @ https://github.com/angular/universal/tree/main/modules/express-engine)
  server.engine("html", ngExpressEngine({
    bootstrap: AppServerModule,
  }));

  server.set("view engine", "html");
  server.set("views", distFolder);

  environment.showDebugInformation = showLogs;

  if (!environment.production) {
    const proxyConfig = proxyConfLocal.obj(undefined);
    Object.keys(proxyConfig).forEach(key => {
      server.use(key, createProxyMiddleware(proxyConfig[key]));
    });
  }

  server.get(staticRoutes, (req, res) => {
    log("server.ts: serve static file : " + req.url);
    let url = req.url;
    const indexOfQueryParam = url.indexOf("?");
    if (indexOfQueryParam !== -1) {
      url = url.substring(0, indexOfQueryParam);
      if (!environment.production) {
        log("server.ts: clean url : " + req.url + " => " + url);
      }
    }
    const start = Date.now();
    return res.sendFile(distFolder + "/" + url, (err: Error) => {
      window[SsrUtil.IS_SSR_FIRST_CALL_PROPERTY_KEY] = false;
      if (isNullOrUndefined(err)) {
        const end = Date.now();
        if (logTime) {
          log(`server.ts: execution time for static file '${req.url}' : ${end - start} ms`);
        }
      }
    });
  });

  // All regular routes use the Universal engine
  server.get(ssrRoutes, (req, res, next) => {
    log("\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> server.ts: run ssr computed page for url : " + req.url);
    const start = Date.now();
    res.render(indexHtml, {
      req,
      providers: [
        {provide: APP_BASE_HREF, useValue: req.baseUrl},
      ],
    }, (err, html) => {
      if (err) {
        log("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< server.ts: error during running SSR page for url : " + req.url, err, html);
        res.sendStatus(500);
        next(err);
      } else {
        const end = Date.now();
        if (logTime) {
          log(`<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< server.ts: execution time for ssr page '${req.url}' : ${end - start} ms\n`);
        }
        res.status(200).send(html);
      }
    });
  });

  server.get("*", (req, res) => {
    log("server.ts: request return default index.html file for url : " + req.url);
    return res.sendFile(distFolder + "/index.html");
  });

  return server;
};

const run = (): void => {
  const port = process.env["PORT"] || DEFAULT_SSR_PORT;
  const showLogs = process.env["SHOW_LOGS"] || DEFAULT_SHOW_LOGS;
  const logTime = process.env["LOG_TIME"] || DEFAULT_LOG_TIME;

  window[SsrUtil.PORT_PROPERTY_KEY] = port;

  // Start up the Node server
  const server = app(!!showLogs, !!logTime);
  server.listen(port, () => {
    // eslint-disable-next-line no-console
    console.log(`Node Express server listening on http://localhost:${port} with param SHOW_LOGS=${showLogs} and LOG_TIME=${logTime}`);
  });
};

// Webpack will replace 'require' with '__webpack_require__'
// '__non_webpack_require__' is a proxy to Node 'require'
// The below code is to ensure that the server is run only when not requiring the bundle.
// eslint-disable-next-line @typescript-eslint/naming-convention
declare const __non_webpack_require__: NodeRequire;
const mainModule = __non_webpack_require__.main;
const moduleFilename = mainModule && mainModule.filename || "";
if (moduleFilename === __filename || moduleFilename.includes("iisnode")) {
  run();
}

export * from "./src/main.server";
